RE_TOLERANCE = 1e-10
TE_TOLERANCE = 1e-10

def valid_dim(input_dim: int) -> bool:
    valid_dims = [1, 2, 3]

    if input_dim not in valid_dims:
        raise ValueError(f'Input value for dim argument ({input_dim}) is not valid. Use one of the following: {valid_dims}.')
        
    else:
        return True

VALID_ROTATION_TYPES = ['euler', 'quaternion', 'angle-axis', 'matrix', 'rodrigues']



import toml
import numpy as np

from .utils import VALID_ROTATION_TYPES
from .pose import Pose
from .transform import Transform


class TransformSet:
    def __init__(self, cfg_file: str) -> None:

        self.__frame_data = toml.load(cfg_file)
        self.frames = dict()

        # Add base frame if not present
        if 'base' not in self.frame_names():
            base_frame = Pose(name='base')
            base_frame.position.zero()
            base_frame.orientation.identity()
            self.frames['base'] = base_frame
        
        # Create dictionary of frames (from which we can create transformations)
        for frame_name, frame_data in self.__frame_data.items():
            self.add_frame(frame_name=frame_name, frame_data=frame_data)

    # Setter functions
    def add_frame(self, frame_name: str, frame_data: dict) -> None:
        new_frame = Pose(name=frame_name)

        # Extract position
        new_frame.position.from_vector(frame_data['position'])

        # Extract orientation
        orientation_value = frame_data['orientation']
        orientation_type = frame_data['orientation_type'].lower()
        degrees = 'degree' in frame_data['orientation_untis'].lower()

        if orientation_type == 'euler':
            new_frame.orientation.from_euler('xyz', orientation_value, degrees=degrees)
        elif orientation_type == 'quaternion':
            new_frame.rotation.from_quat(orientation_value)
        elif orientation_type == 'angle-axis':
            new_frame.rotation.from_angle_axis(orientation_value)
        elif orientation_type == 'matrix':
            new_frame.rotation.from_matrix(orientation_value)
        else:
            raise ValueError(f'TransformSet - Invalid rotation type: {orientation_type}. Rotation type must be: {VALID_ROTATION_TYPES}')

        # Save new frame to self.frames
        self.frames[frame_name] = new_frame


    # Getter functions
    def frame_names(self) -> list:
        return self.frames.keys()


    def change_frame(self, input, from_frame: str, to_frame: str) -> np.ndarray:
        '''
        Coordinate transformation of a pose (6D vector) from origin frame to target frame.

        A compund transformation from origin frame (defined in `from_frame` argument) to
        the target frame (defined in `to_frame` argument) is computed and applied to the
        input pose.

        Args:
            input (np.ndarray): Input pose
            from_frame (str): Name of origin frame
            to_frame (str): Name of target frame

        Returns:
            np.ndarray: Transformed pose in target frame.
        '''
        # Create compound transformation
        transformation = self.__create_compound_transf(from_frame=from_frame, to_frame=to_frame)

        return transformation.apply(input)


    def wrench_change_frame(self, wrench: np.ndarray, from_frame: str, to_frame: str) -> np.ndarray:
        '''
        Function to change frame of wrench vector.

        Function will perform simple rotation on forces (first three elements), and
        will rotate the total moments on the origin frame.

        Args:
            wrench (np.ndarray): Input wrench array
            from_frame (str): Name of origin frame
            to_frame (str): Name of target frame

        Returns:
            np.ndarray: Transformed wrench array
        '''
        # Verify input
        if not np.array(wrench).shape == (6,):
            raise ValueError(f"TransformSet - Invalid wrench input. Shape must be (6,)")

        # Create compound transformation
        transformation = self.__create_compound_transf(from_frame=from_frame, to_frame=to_frame)

        # Transform wrench
        force_at_orig = wrench[:3]
        torque_at_orig = wrench[3:]

        torque_at_dest = transformation.rotation.apply(np.cross(force_at_orig, transformation.translation) + torque_at_orig)
        force_at_dest = transformation.rotation.apply(force_at_orig)

        return np.hstack([force_at_dest, torque_at_dest])


    def transform_matrix(self, from_frame: str, to_frame: str, homogeneous: bool = True) -> np.ndarray:
        '''
        Return the transformation matrix to transform poses from origin
        frame to destination frame.

        Function will call the `__create_compound_transf()` function. Note that such a matrix
        can only be directly used for poses. Other calculations are required for wrench
        transformations.

        Args:
            from_frame (str): Name of origin frame
            to_frame (str): Name of target frame
            homogeneous (bool, optional): Option if matrix should be homogenous or not (3x4 or 4x4). Defaults to True.

        Returns:
            np.ndarray: Numpy matrix
        '''
        # Create compound transformation
        full_transf = self.__create_compound_transf(from_frame, to_frame)

        return full_transf.matrix(homogeneous=homogeneous)


    def __create_compound_transf(self, from_frame: str, to_frame: str) -> Transform:
        '''
        Function to create compound transform between two frames.

        The function will create a transformation from one frame to another by 

        Args:
            from_frame (str): Name of origin frame
            to_frame (str): Name of destination frame

        Returns:
            Transform: Transform object
        '''
        transformation = Transform(name=f'{from_frame}2{to_frame}', orig=from_frame, dest=to_frame)
        transformation.between_poses(pose_1=self.frames[from_frame], pose_2=self.frames[to_frame])

        return transformation



import numpy as np

from .utils import valid_dim

class TE:
    def __init__(self, name: str ='', dim: int = 3, vector: np.ndarray = None) -> None:
        self.name = name

        if valid_dim(dim):
            self.__dim = dim

        if vector is None:
            self.__vector = np.zeros(self.__dim)
        else:
            if valid_dim(len(vector)):
                self.__dim = len(vector)
                self.__vector = vector

    # Setter functions
    def random(self) -> None:
        self.from_vector(np.random.rand(self.__dim))

    def from_vector(self, vector: np.ndarray) -> None:
        if vector.shape == self.__vector.shape:
            self.__vector = vector

    def zero(self) -> None:
        self.from_vector(np.zeros(self.__dim))

    def inv(self) -> None:
        self.from_vector(-self.vector())

    # Getter functions
    def vector(self) -> np.ndarray:
        return self.__vector

    def x(self) -> float:
        return float(self.vector()[0])

    def y(self) -> float:
        return float(self.vector()[1])

    def z(self) -> float:
        return float(self.vector()[2])

    # Operator overloads
    def __str__(self) -> str:
        return f'TE{self.__dim} - {self.name}: {self.vector()}'

    def __repr__(self) -> str:
        return f'{self.vector()}'

    def __add__(self, other):
        if isinstance(other, TE):
            if other.vector().shape == self.vector().shape:
                return TE(name=f'Sum of {self.name} and {other.name}',
                          vector=self.vector() + other.vector())

        elif isinstance(other, np.ndarray):
            if other.shape == self.vector().shape:
                return TE(name=self.name,
                           vector=self.vector() + other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __sub__(self, other):
        if isinstance(other, TE):
            if other.vector().shape == self.vector().shape:
                return TE(name=f'Sum of {self.name} and {other.name}',
                          vector=self.vector() - other.vector())

        elif isinstance(other, np.ndarray):
            if other.shape == self.vector().shape:
                return TE(name=self.name,
                           vector=self.vector() - other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __iadd__(self, other):
        if isinstance(other, TE):
            if other.vector().shape == self.vector().shape:
                self.from_vector(self.vector() + other.vector())

        elif isinstance(other, np.ndarray):
            if other.shape == self.vector().shape:
                self.from_vector(self.vector() + other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __isub__(self, other):
        if isinstance(other, TE):
            if other.vector().shape == self.vector().shape:
                self.from_vector(self.vector() - other.vector())

        elif isinstance(other, np.ndarray):
            if other.shape == self.vector().shape:
                self.from_vector(self.vector() - other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __eq__(self, other):
        if isinstance(other, TE):
            return np.array_equal(self.vector(), other.vector())
            
        elif isinstance(other, np.ndarray):
            return np.array_equal(self.vector(), other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __ne__(self, other):
        if isinstance(other, TE):
            return not np.array_equal(self.vector(), other.vector())
            
        elif isinstance(other, np.ndarray):
            return not np.array_equal(self.vector(), other)

        else:
            raise TypeError(f'Input parameter is {type(other)}, not TE or np.ndarray as expected.')

    def __neg__(self):
        self.from_vector(-self.vector())

    def __abs__(self):
        self.from_vector(abs(self.vector()))



import numpy as np

from .te import TE
from .re import RE

from .pose import Pose

from .utils import valid_dim

class Transform:
    def __init__(self, name: str, orig: str = 'origin', dest: str = 'destination', dim: int = 3) -> None:
        # Set strings
        self.name = name
        self.origin = orig
        self.destination = dest

        # Init translation and orientation
        if valid_dim(dim):
            self.__dim = dim
            self.translation = TE(dim=dim)
            self.rotation = RE(dim=dim)

    # Setter functions
    def between_poses(self, pose_1: Pose, pose_2: Pose):
        '''
        Compute transform between 2 poses (3D).

        This instance of Transform will be modifed to compute the transform from pose_1 to pose_2.

        Args:
            pose_1 (Pose): Origin pose.
            pose_2 (Pose): Destination pose.
        '''
        if pose_1.__dim != pose_2.__dim:
            raise AttributeError(f'Number of dimensions between both poses do not match: pose_1.__dim = {pose_1.__dim} and pose_2.__dim = {pose_2.__dim}')

        # Modify dimension of transformation depending on pose_1 and pose_2
        if pose_1.__dim != self.__dim:
            self.__dim = pose_1.__dim
            self.rotation = RE(dim=self.__dim)

        # Compute rotation from pose_1 to pose_2
        self.rotation.from_matrix(np.divide(pose_2.orientation.as_matrix(), pose_1.orientation.as_matrix()))
        self.translation.from_vector(pose_2.position.vector() - pose_1.position.vector())


    def identity(self) -> None:
        self.translation.zero()
        self.rotation.identity()
        

    def inv(self):
        self.rotation = self.rotation.inv()
        self.translation.from_vector(-self.rotation.apply(self.translation.vector()))

    def random(self):
        self.translation.random()
        self.rotation.random()
    
    # Getter functions
    def matrix(self, homogeneous: bool = True) -> np.ndarray:

        matrix = np.eye(self.__dim + 1)
        
        if self.orig != self.dest:
            matrix[:self.__dim, :self.__dim] = self.rotation.as_matrix()
            matrix[:self.__dim, self.__dim] = self.translation

        if not homogeneous:
            return matrix[:self.__dim, :]
        
        return matrix

    # Computation functions
    def apply(self, io):
        # If io is pose
        if isinstance(io, Pose):
            io.orientation.from_matrix(np.matmul(self.rotation.as_matrix(), io.orientation.as_matrix()))
            io.position += self.translation

        # If io is numpy vector
        if isinstance(io, np.ndarray):
            io = self.rotation.apply(io) + self.translation

        return io

    # Operator overloads
    def __repr__(self) -> str:
        return f'''Transform ({self.__dim}D) - {self.name}:
        Position:    {self.position.__repr__}
        Orientation: {self.orientation.__repr__}'''

    def __str__(self) -> str:
        return f'Translation: {self.position.__repr__}\nRotation:    {self.orientation.__repr__}'

    def __eq__(self, other: object) -> bool:
        return self.translation == other.translation and self.rotation == other.rotation

    def __ne__(self, other: object) -> bool:
        return self.translation != other.translation or self.rotation != other.rotation


from pathlib import Path
from setuptools import setup, find_packages

VERSION = '2.0.1'
DESCRIPTION = 'Transforming and handling 3D poses and frames.'

# Read the contents of README file
this_directory = Path(__file__).parent
requirements_file = this_directory / 'requirements.txt'
long_description = (this_directory / "README.md").read_text()

# Add resource links
project_urls = {
    'Documentation': 'https://johnhal.gitlab.io/pose_python',
    'Repository': 'https://gitlab.com/johnhal/pose3d_python'
}

# Setting up
setup(
    name="pose3d",
    version=VERSION,
    author="John Halazonetis",
    author_email="<john.halazonetis@icloud.com>",
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=list(filter(None, open(requirements_file.as_posix()).read().split('\n'))),
    keywords=['python', 'pose', 'transform'],
    project_urls=project_urls,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from .te import TE
from .re import RE

from .utils import valid_dim

class Pose:
    def __init__(self, name: str = '', dim: int = 3) -> None:
        self.name = name

        if valid_dim(dim):
            self.__dim = dim
            self.position = TE(dim=dim)
            self.orientation = RE(dim=dim)

    def random(self):
        self.orientation.random()
        self.position.random()

    def zero(self):
        self.orientation.identity()
        self.position.zero()

    # Operator overloads
    def __str__(self) -> str:
        return f'''Pose ({self.__dim}D) - {self.name}:
        Position:    {self.position.__repr__}
        Orientation: {self.orientation.__repr__}'''

    def __repr__(self) -> str:
        return f'Position:    {self.position.__repr__}\nOrientation: {self.orientation.__repr__}'

    def __eq__(self, other) -> bool:
        if isinstance(other, Pose) and other.__dim == self.__dim:
            return self.orientation == other.orientation and self.position == other.position
        else:
            return False

    def __ne__(self, other) -> bool:
        if isinstance(other, Pose) and other.__dim == self.__dim:
            return self.orientation != other.orientation or self.position != other.position
        else:
            return False


import numpy as np
from scipy.spatial.transform import Rotation
from .utils import RE_TOLERANCE, valid_dim

class RE:
    def __init__(self, name: str = '', dim: int = 3) -> None:
        self.name = name
        self.__rotation = Rotation(quat=[0, 0, 0, 1])

        if valid_dim(dim):
            self.__dim = dim

    # Setter functions
    def identity(self):
        self.__rotation = Rotation.identity()

    def inv(self):
        self.__rotation = self.__rotation.inv()

    def random(self):
        self.__rotation = Rotation.random()

    def from_quat(self, quat: np.ndarray) -> None:
        if self.__dim == 2:
            raise AttributeError(f'Unable to set 2D rotation from quaternion input.')

        self.__rotation = Rotation.from_quat(quat)

    def from_matrix(self, matrix: np.ndarray) -> None:
        if matrix.shape != (self.__dim, self.__dim):
            raise ValueError(f'Input matrix shape must be ({self.__dim}, {self.__dim}) when rotation dimension is {self.__dim}. Current input matrix shape: {matrix.shape}.')

        if self.__dim == 2:
            matrix = np.hstack(matrix, np.zeros(2))
            matrix = np.vstack(matrix, [0, 0, 1])
        
        self.__rotation = Rotation.from_matrix(matrix)

    def from_angle_axis(self, angle_axis: np.ndarray) -> None:
        if self.__dim == 2:
            raise AttributeError(f'Unable to set 2D rotation from angle-axis input.')

        self.__rotation = Rotation.from_rotvec(angle_axis)

    def from_euler(self, sequence: str = None, angles: list = None, degrees: bool = True) -> None:
        if self.__dim == 3:
            self.__rotation = Rotation.from_euler(sequence, angles, degrees)

        # TODO: Verify functionality with test
        elif self.__dim == 2:
            self.__rotation = Rotation.from_euler('z', angles, degrees)

    # Getter functions
    def as_quat(self) -> np.ndarray:
        return self.__rotation.as_quat()

    def as_matrix(self) -> np.ndarray:
        return self.__rotation.as_matrix()[:self.__dim, :self.__dim]

    def as_angle_axis(self) -> np.ndarray:
        return self.__rotation.as_rotvec()

    def as_euler(self, sequence: str = None, degrees: bool = True):
        if self.__dim == 3:
            return self.__rotation.as_euler(sequence, degrees)

        elif self.__dim == 2:
            return self.__rotation.as_euler('z', degrees)[0]

    def yaw(self, degrees: bool = True) -> float:
        '''
        Return rotation angle around the z axis

        Args:
            degrees (bool, optional): Option to return value in degrees or radians. Defaults to True.

        Returns:
            float: Yaw angle (in specified units)
        '''
        if self.__dim == 2:
            raise AttributeError(f'Unable to return yaw angle of 2D rotation (Call as_euler instead).')

        return self.__rotation.as_euler('z', degrees)[0]

    def pitch(self, degrees: bool = True) -> float:
        '''
        Return rotation angle around the y axis

        Args:
            degrees (bool, optional): Option to return value in degrees or radians. Defaults to True.

        Returns:
            float: Pitch angle (in specified units)
        '''
        if self.__dim == 2:
            raise AttributeError(f'Unable to return pitch angle of 2D rotation (Call as_euler instead).')

        return self.__rotation.as_euler('y', degrees)[0]

    def roll(self, degrees: bool = True) -> float:
        '''
        Return rotation angle around the x axis

        Args:
            degrees (bool, optional): Option to return value in degrees or radians. Defaults to True.

        Returns:
            float: Roll angle (in specified units)
        '''
        if self.__dim == 2:
            raise AttributeError(f'Unable to return roll angle of 2D rotation (Call as_euler instead).')

        return self.__rotation.as_euler('x', degrees)[0]

    # Computation functions
    def apply(self, input):
        # Check shape of input
        if input.shape[0] != self.__dim:
            raise ValueError(f'Input shape mismatch: self.__dim ({self.__dim}) != input.shape ({input.shape[0]})')
        
        
        return self.__rotation.apply(input)

    # Operator overloading
    def __str__(self) -> str:
        return f'RE - {self.name}: {self.__repr__()} degrees'

    def __repr__(self) -> str:
        if self.__dim == 3:
            sequence = 'xyz'
        if self.__dim == 2:
            sequence = 'z'

        return f'{self.as_euler(sequence, True)} degrees'

    def __eq__(self, other):
        if isinstance(other, RE):
            return np.allclose(self.as_quat(),
                               other.as_quat(),
                               rtol=RE_TOLERANCE,
                               atol=RE_TOLERANCE)
        else:
            raise TypeError(f'Input parameter is {type(other)}, not RE as expected.')

    def __ne__(self, other):
        if isinstance(other, RE):
            return not np.allclose(self.as_quat(),
                                   other.as_quat(),
                                   rtol=RE_TOLERANCE,
                                   atol=RE_TOLERANCE)
        else:
            raise TypeError(f'Input parameter is {type(other)}, not RE as expected.')
    